/*
 * Transparent proxy support for Linux/iptables
 *
 * Copyright (C) 2007 BalaBit IT Ltd.
 * Author: Krisztian Kovacs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include <net/inet_sock.h>
#include <net/netfilter/nf_tproxy_core.h>
#include <net/netfilter/ipv4/nf_defrag_ipv4.h>

#if defined(CONFIG_NF_CONNTRACK) || defined(CONFIG_NF_CONNTRACK_MODULE)
#define XT_SOCKET_HAVE_CONNTRACK 1
#include <net/netfilter/nf_conntrack.h>
#endif

static bool
socket_mt(const struct sk_buff *skb,
	  const struct net_device *in,
	  const struct net_device *out,
	  const struct xt_match *match,
	  const void *matchinfo,
	  int offset,
	  unsigned int protoff,
	  bool *hotdrop)
{
	const struct iphdr *iph = ip_hdr(skb);
	struct udphdr _hdr, *hp;
	struct sock *sk;
	__be32 daddr;
	__be16 dport;
#ifdef XT_SOCKET_HAVE_CONNTRACK
	struct nf_conn const *ct;
	enum ip_conntrack_info ctinfo;
#endif

	hp = skb_header_pointer(skb, ip_hdrlen(skb), sizeof(_hdr), &_hdr);
	if (hp == NULL)
		return false;

	daddr = iph->daddr;
	dport = hp->dest;

#ifdef XT_SOCKET_HAVE_CONNTRACK
	/* Do the lookup with the original socket address in case this is a 
	 * reply packet of an established SNAT-ted connection. */
	ct = nf_ct_get(skb, &ctinfo);
	if (ct && (ct != &nf_conntrack_untracked) &&
	    (ctinfo == IP_CT_IS_REPLY + IP_CT_ESTABLISHED) &&
	    (ct->status & IPS_SRC_NAT_DONE)) {
		daddr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u3.ip;
		dport = (iph->protocol == IPPROTO_TCP) ?
			ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.tcp.port :
			ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	}
#endif

	sk = nf_tproxy_get_sock_v4(skb->dev->nd_net, iph->protocol,
				   iph->saddr, daddr,
				   hp->source, dport, in, false);
	if (sk != NULL) {
		bool wildcard = (inet_sk(sk)->rcv_saddr == 0);

		nf_tproxy_put_sock(sk);
		if (wildcard)
			sk = NULL;
	}

	pr_debug("socket match: proto %u %08x:%u -> %08x:%u (orig %08x:%u) sock %p\n",
		 iph->protocol, ntohl(iph->saddr), ntohs(hp->source),
		 ntohl(daddr), ntohs(dport),
		 ntohl(iph->daddr), ntohs(hp->dest), sk);

	return (sk != NULL);
}

static bool
socket_mt_check(const char *tablename,
		const void *entry,
		const struct xt_match *match,
		void *matchinfo,
		unsigned int hook_mask)
{
	const struct ipt_ip *i = entry;

	if ((i->proto == IPPROTO_TCP || i->proto == IPPROTO_UDP)
	    && !(i->invflags & IPT_INV_PROTO))
		return true;

	pr_info("xt_socket: Can be used only in combination with "
		"either -p tcp or -p udp\n");
	return false;
}

static struct xt_match socket_mt_reg __read_mostly = {
	.name		= "socket",
	.family		= AF_INET,
	.match		= socket_mt,
	.checkentry	= socket_mt_check,
	.hooks		= 1 << NF_INET_PRE_ROUTING,
	.me		= THIS_MODULE,
};

static int __init socket_mt_init(void)
{
	nf_defrag_ipv4_enable();
	return xt_register_match(&socket_mt_reg);
}

static void __exit socket_mt_exit(void)
{
	xt_unregister_match(&socket_mt_reg);
}

module_init(socket_mt_init);
module_exit(socket_mt_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Krisztian Kovacs");
MODULE_DESCRIPTION("x_tables socket match module");
MODULE_ALIAS("ipt_socket");
